@@ -1,11 +1,13 @@
+# pylint:disable=missing-class-docstring
 import logging
 
 import pyvex
 from angr.utils.constants import DEFAULT_STATEMENT
 from angr.engines.vex.claripy.irop import vexop_to_simop
+from angr.errors import UnsupportedIROpError
 
 from .block import Block
-from .statement import Assignment, Store, Jump, Call, ConditionalJump, DirtyStatement, Return
+from .statement import Assignment, CAS, Store, Jump, Call, ConditionalJump, DirtyStatement, Return
 from .expression import (
     Const,
     Register,
@@ -18,7 +20,6 @@ from .expression import (
     ITE,
     Reinterpret,
     VEXCCallExpression,
-    TernaryOp,
 )
 from .converter_common import SkipConversionNotice, Converter
 
@@ -42,35 +43,27 @@ class VEXExprConverter(Converter):
         :param expr:
         :return:
         """
-        func = EXPRESSION_MAPPINGS.get(type(expr))
-        if func is not None:
-            return func(expr, manager)
-
         if isinstance(expr, pyvex.const.IRConst):
             return VEXExprConverter.const_n(expr, manager)
 
-        if isinstance(expr, pyvex.IRExpr.CCall):
-            operands = tuple(VEXExprConverter.convert(arg, manager) for arg in expr.args)
-            ccall = VEXCCallExpression(
-                manager.next_atom(),
-                expr.cee.name,
-                operands,
-                bits=expr.result_size(manager.tyenv),
-                ins_addr=manager.ins_addr,
-                vex_block_addr=manager.block_addr,
-                vex_stmt_idx=manager.vex_stmt_idx,
-            )
-            return DirtyExpression(
-                manager.next_atom(),
-                ccall,
-                bits=expr.result_size(manager.tyenv),
-                ins_addr=manager.ins_addr,
-                vex_block_addr=manager.block_addr,
-                vex_stmt_idx=manager.vex_stmt_idx,
-            )
+        func = EXPRESSION_MAPPINGS.get(type(expr))
+        if func is not None:
+            # When something goes wrong, return a DirtyExpression instead of crashing the program
+            try:
+                return func(expr, manager)
+            except UnsupportedIROpError:
+                log.warning("VEXExprConverter: Unsupported IROp %s.", expr.op)
+                return DirtyExpression(
+                    manager.next_atom(), f"unsupported_{expr.op}", [], bits=expr.result_size(manager.tyenv)
+                )
 
         log.warning("VEXExprConverter: Unsupported VEX expression of type %s.", type(expr))
-        return DirtyExpression(manager.next_atom(), expr, bits=expr.result_size(manager.tyenv))
+        try:
+            bits = expr.result_size(manager.tyenv)
+        except ValueError:
+            # e.g., "ValueError: Type Ity_INVALID does not have size"
+            bits = 0
+        return DirtyExpression(manager.next_atom(), f"unsupported_{str(type(expr))}", [], bits=bits)
 
     @staticmethod
     def convert_list(exprs, manager):
@@ -152,7 +145,18 @@ class VEXExprConverter(Converter):
                     shifted = BinaryOp(
                         manager.next_atom(),
                         "Shr",
-                        [inner, Const(manager.next_atom(), None, simop._to_size, 8)],
+                        [
+                            inner,
+                            Const(
+                                manager.next_atom(),
+                                None,
+                                simop._to_size,
+                                8,
+                                ins_addr=manager.ins_addr,
+                                vex_block_addr=manager.block_addr,
+                                vex_stmt_idx=manager.vex_stmt_idx,
+                            ),
+                        ],
                         False,
                         ins_addr=manager.ins_addr,
                         vex_block_addr=manager.block_addr,
@@ -180,11 +184,15 @@ class VEXExprConverter(Converter):
                     vex_stmt_idx=manager.vex_stmt_idx,
                 )
             raise NotImplementedError("Unsupported operation")
+        elif op_name == "Not" and expr.op != "Iop_Not1":
+            # NotN (N != 1) is equivalent to bitwise negation
+            op_name = "BitwiseNeg"
 
         return UnaryOp(
             manager.next_atom(),
             op_name,
             VEXExprConverter.convert(expr.args[0], manager),
+            bits=expr.result_size(manager.tyenv),
             ins_addr=manager.ins_addr,
             vex_block_addr=manager.block_addr,
             vex_stmt_idx=manager.vex_stmt_idx,
@@ -203,9 +211,17 @@ class VEXExprConverter(Converter):
             operands[1] = Const(operands[1].idx, None, (1 << op1_bits) - op1_val, op1_bits)
 
         signed = False
-        if op_name in {"CmpLE", "CmpLT", "CmpGE", "CmpGT", "Div", "DivMod", "Mod", "Mul", "Mull"}:
+        vector_count = None
+        vector_size = None
+        if op._vector_count is not None and op._vector_size is not None:
+            # SIMD conversions
+            op_name += "V"  # vectorized
+            vector_count = op._vector_count
+            vector_size = op._vector_size
+        elif op_name in {"CmpLE", "CmpLT", "CmpGE", "CmpGT", "Div", "DivMod", "Mod", "Mul", "Mull"}:
             if op.is_signed:
                 signed = True
+
         if op_name == "Cmp" and op._float:
             # Rename Cmp to CmpF
             op_name = "CmpF"
@@ -271,6 +287,69 @@ class VEXExprConverter(Converter):
 
         bits = op._output_size_bits
 
+        if op_name == "DivMod":
+            op1_size = op._from_size if op._from_size is not None else operands[0].bits
+            op2_size = op._to_size if op._to_size is not None else operands[1].bits
+
+            if op2_size < op1_size:
+                # e.g., DivModU64to32
+                operands[1] = Convert(
+                    manager.next_atom(),
+                    op2_size,
+                    op1_size,
+                    op._from_signed != "U",
+                    operands[1],
+                    ins_addr=manager.ins_addr,
+                    vex_block_addr=manager.block_addr,
+                    vex_stmt_idx=manager.vex_stmt_idx,
+                )
+            chunk_bits = bits // 2
+
+            div = BinaryOp(
+                manager.next_atom(),
+                "Div",
+                operands,
+                signed,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+                bits=op1_size,
+            )
+            truncated_div = Convert(
+                manager.next_atom(),
+                op1_size,
+                chunk_bits,
+                signed,
+                div,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+            )
+            mod = BinaryOp(
+                manager.next_atom(),
+                "Mod",
+                operands,
+                signed,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+                bits=op1_size,
+            )
+            truncated_mod = Convert(
+                manager.next_atom(),
+                op1_size,
+                chunk_bits,
+                signed,
+                mod,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+            )
+
+            operands = [truncated_mod, truncated_div]
+            op_name = "Concat"
+            signed = False
+
         return BinaryOp(
             manager.next_atom(),
             op_name,
@@ -280,6 +359,8 @@ class VEXExprConverter(Converter):
             vex_block_addr=manager.block_addr,
             vex_stmt_idx=manager.vex_stmt_idx,
             bits=bits,
+            vector_count=vector_count,
+            vector_size=vector_size,
         )
 
     @staticmethod
@@ -307,14 +388,9 @@ class VEXExprConverter(Converter):
                 bits=bits,
             )
 
-        return TernaryOp(
-            manager.next_atom(),
-            op_name,
-            operands,
-            ins_addr=manager.ins_addr,
-            vex_block_addr=manager.block_addr,
-            vex_stmt_idx=manager.vex_stmt_idx,
-            bits=bits,
+        raise TypeError(
+            "Please figure out what kind of operation this is (smart money says fused multiply) and convert it into "
+            "multiple binops"
         )
 
     @staticmethod
@@ -359,6 +435,19 @@ class VEXExprConverter(Converter):
             vex_stmt_idx=manager.vex_stmt_idx,
         )
 
+    @staticmethod
+    def CCall(expr: pyvex.IRExpr.CCall, manager):
+        operands = [VEXExprConverter.convert(arg, manager) for arg in expr.args]
+        return VEXCCallExpression(
+            manager.next_atom(),
+            expr.cee.name,
+            operands,
+            bits=expr.result_size(manager.tyenv),
+            ins_addr=manager.ins_addr,
+            vex_block_addr=manager.block_addr,
+            vex_stmt_idx=manager.vex_stmt_idx,
+        )
+
 
 EXPRESSION_MAPPINGS = {
     pyvex.IRExpr.RdTmp: VEXExprConverter.RdTmp,
@@ -371,6 +460,7 @@ EXPRESSION_MAPPINGS = {
     pyvex.const.U64: VEXExprConverter.const_n,
     pyvex.IRExpr.Load: VEXExprConverter.Load,
     pyvex.IRExpr.ITE: VEXExprConverter.ITE,
+    pyvex.IRExpr.CCall: VEXExprConverter.CCall,
 }
 
 
@@ -388,7 +478,8 @@ class VEXStmtConverter(Converter):
         try:
             func = STATEMENT_MAPPINGS[type(stmt)]
         except KeyError:
-            return DirtyStatement(idx, stmt, ins_addr=manager.ins_addr)
+            dirty = DirtyExpression(manager.next_atom(), str(stmt), [], bits=0)
+            return DirtyStatement(idx, dirty, ins_addr=manager.ins_addr)
 
         return func(idx, stmt, manager)
 
@@ -468,7 +559,7 @@ class VEXStmtConverter(Converter):
             "ILGop_16Sto32": (16, 32, True),
         }
 
-        dst = VEXExprConverter.tmp(stmt.dst, manager.tyenv.sizeof(stmt.dst) // 8, manager)
+        dst = VEXExprConverter.tmp(stmt.dst, manager.tyenv.sizeof(stmt.dst), manager)
         load_bits, convert_bits, signed = sizes[stmt.cvt]
         src = Load(
             manager.next_atom(),
@@ -506,96 +597,59 @@ class VEXStmtConverter(Converter):
 
     @staticmethod
     def CAS(idx, stmt: pyvex.IRStmt.CAS, manager):
-        # compare-and-swap is translated into multiple statements. the atomic property is lost.
-
-        stmts = []
-
-        double = stmt.dataHi is not None
-        ty = stmt.expdLo.result_type(manager.tyenv)
-        if double:
-            ty, narrow_to_bits, widen_from_bits, widen_to_bits = {
-                "Ity_I8": ("Ity_I16", 8, 8, 16),
-                "Ity_I16": ("Ity_I32", 16, 16, 32),
-                "Ity_I32": ("Ity_I64", 32, 32, 64),
-                "Ity_I64": ("Ity_V128", 64, 64, 128),
-            }[ty]
-            dataHi = VEXExprConverter.convert(stmt.dataHi, manager)
-            dataLo = VEXExprConverter.convert(stmt.dataLo, manager)
-            data = BinaryOp(idx, "Concat", (dataHi, dataLo), False)
-
-            expdHi = Convert(idx, widen_from_bits, widen_to_bits, False, VEXExprConverter.convert(stmt.dataHi, manager))
-            expdLo = Convert(idx, widen_from_bits, widen_to_bits, False, VEXExprConverter.convert(stmt.dataLo, manager))
-            expd = BinaryOp(idx, "Concat", (expdHi, expdLo), False)
-        else:
-            narrow_to_bits = widen_to_bits = None
-            data = VEXExprConverter.convert(stmt.dataLo, manager)
-            expd = VEXExprConverter.convert(stmt.expdLo, manager)
-
-        size = {
-            "Ity_I8": 1,
-            "Ity_I16": 2,
-            "Ity_I32": 4,
-            "Ity_I64": 8,
-            "Ity_V128": 16,
-        }[ty]
-
-        # load value from memory
+        # addr
         addr = VEXExprConverter.convert(stmt.addr, manager)
-        val = Load(
-            idx,
-            addr,
-            size,
-            stmt.endness,
+        data_lo = VEXExprConverter.convert(stmt.dataLo, manager)
+        data_hi = VEXExprConverter.convert(stmt.dataHi, manager) if stmt.dataHi is not None else None
+        expd_lo = VEXExprConverter.convert(stmt.expdLo, manager)
+        expd_hi = VEXExprConverter.convert(stmt.expdHi, manager) if stmt.expdHi is not None else None
+        old_lo = VEXExprConverter.tmp(stmt.oldLo, manager.tyenv.sizeof(stmt.oldLo), manager)
+        old_hi = (
+            VEXExprConverter.tmp(stmt.oldHi, stmt.oldHi.result_size(manager.tyenv), manager)
+            if stmt.oldHi != 0xFFFFFFFF
+            else None
         )
-        cmp = BinaryOp(idx, "CmpEQ", (val, expd), False)
-        store = Store(
+        return CAS(
+            idx, addr, data_lo, data_hi, expd_lo, expd_hi, old_lo, old_hi, stmt.endness, ins_addr=manager.ins_addr
+        )
+
+    @staticmethod
+    def Dirty(idx, stmt: pyvex.IRStmt.Dirty, manager):
+        # we translate it into tmp = DirtyExpression() if possible
+
+        operands = [VEXExprConverter.convert(op, manager) for op in stmt.args]
+        guard = VEXExprConverter.convert(stmt.guard, manager) if stmt.guard is not None else None
+        bits = manager.tyenv.sizeof(stmt.tmp) if stmt.tmp != 0xFFFFFFFF else 0
+        maddr = VEXExprConverter.convert(stmt.mAddr, manager) if stmt.mAddr is not None else None
+        dirty_expr = DirtyExpression(
+            manager.next_atom(),
+            stmt.cee.name,
+            operands,
+            guard=guard,
+            mfx=stmt.mFx,
+            maddr=maddr,
+            msize=stmt.mSize,
+            bits=bits,
+        )
+
+        if stmt.tmp == 0xFFFFFFFF:
+            return DirtyStatement(
+                idx,
+                dirty_expr,
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=manager.vex_stmt_idx,
+            )
+
+        tmp = VEXExprConverter.tmp(stmt.tmp, bits, manager)
+        return Assignment(
             idx,
-            addr.copy(),
-            data,
-            size,
-            stmt.endness,
-            guard=cmp,
+            tmp,
+            dirty_expr,
             ins_addr=manager.ins_addr,
             vex_block_addr=manager.block_addr,
             vex_stmt_idx=manager.vex_stmt_idx,
         )
-        stmts.append(store)
-
-        if double:
-            val_shifted = BinaryOp(idx, "Shr", (val, narrow_to_bits), False)
-            valHi = Convert(idx, widen_to_bits, narrow_to_bits, False, val_shifted)
-            valLo = Convert(idx, widen_to_bits, narrow_to_bits, False, val)
-
-            wrtmp_0 = Assignment(
-                idx,
-                Tmp(idx, None, stmt.oldLo, narrow_to_bits),
-                valLo,
-                ins_addr=manager.ins_addr,
-                vex_block_addr=manager.block_addr,
-                vex_stmt_idx=manager.vex_stmt_idx,
-            )
-            wrtmp_1 = Assignment(
-                idx,
-                Tmp(idx, None, stmt.oldHi, narrow_to_bits),
-                valHi,
-                ins_addr=manager.ins_addr,
-                vex_block_addr=manager.block_addr,
-                vex_stmt_idx=manager.vex_stmt_idx,
-            )
-            stmts.append(wrtmp_0)
-            stmts.append(wrtmp_1)
-        else:
-            wrtmp = Assignment(
-                idx,
-                Tmp(idx, None, stmt.oldLo, size),
-                val,
-                ins_addr=manager.ins_addr,
-                vex_block_addr=manager.block_addr,
-                vex_stmt_idx=manager.vex_stmt_idx,
-            )
-            stmts.append(wrtmp)
-
-        return stmts
 
 
 STATEMENT_MAPPINGS = {
@@ -606,6 +660,7 @@ STATEMENT_MAPPINGS = {
     pyvex.IRStmt.StoreG: VEXStmtConverter.StoreG,
     pyvex.IRStmt.LoadG: VEXStmtConverter.LoadG,
     pyvex.IRStmt.CAS: VEXStmtConverter.CAS,
+    pyvex.IRStmt.Dirty: VEXStmtConverter.Dirty,
 }
 
 
@@ -660,9 +715,7 @@ class VEXIRSBConverter(Converter):
                 pass
 
         manager.vex_stmt_idx = DEFAULT_STATEMENT
-        if irsb.jumpkind == "Ijk_Call":
-            # call
-
+        if irsb.jumpkind == "Ijk_Call" or irsb.jumpkind.startswith("Ijk_Sys"):
             # FIXME: Move ret_expr and fp_ret_expr creation into angr because we cannot reliably determine which
             #  expressions can be returned from the call without performing further analysis
             ret_reg_offset = manager.arch.ret_offset
@@ -672,23 +725,37 @@ class VEXIRSBConverter(Converter):
                 ret_reg_offset,
                 manager.arch.bits,
                 reg_name=manager.arch.translate_register_name(ret_reg_offset, size=manager.arch.bits),
+                ins_addr=manager.ins_addr,
+                vex_block_addr=manager.block_addr,
+                vex_stmt_idx=DEFAULT_STATEMENT,
             )
             fp_ret_reg_offset = manager.arch.fp_ret_offset
-            if fp_ret_reg_offset is not None and fp_ret_reg_offset != ret_expr:
+            if fp_ret_reg_offset is not None and fp_ret_reg_offset != ret_reg_offset:
                 fp_ret_expr = Register(
                     manager.next_atom(),
                     None,
                     fp_ret_reg_offset,
                     manager.arch.bits,
                     reg_name=manager.arch.translate_register_name(fp_ret_reg_offset, size=manager.arch.bits),
+                    ins_addr=manager.ins_addr,
+                    vex_block_addr=manager.block_addr,
+                    vex_stmt_idx=DEFAULT_STATEMENT,
                 )
             else:
                 fp_ret_expr = None
 
+            if irsb.jumpkind == "Ijk_Call":
+                target = VEXExprConverter.convert(irsb.next, manager)
+            elif irsb.jumpkind.startswith("Ijk_Sys"):
+                # FIXME: This is a hack to make syscall work. We should have a better way to handle syscalls.
+                target = DirtyExpression(manager.next_atom(), "syscall", [], bits=manager.arch.bits)
+            else:
+                raise NotImplementedError("Unsupported jumpkind")
+
             statements.append(
                 Call(
                     manager.next_atom(),
-                    VEXExprConverter.convert(irsb.next, manager),
+                    target,
                     ret_expr=ret_expr,
                     fp_ret_expr=fp_ret_expr,
                     ins_addr=manager.ins_addr,
@@ -697,9 +764,9 @@ class VEXIRSBConverter(Converter):
                 )
             )
         elif irsb.jumpkind == "Ijk_Boring":
-            if len(conditional_jumps) == 1:
+            if conditional_jumps:
                 # fill in the false target
-                cond_jump = conditional_jumps[0]
+                cond_jump = conditional_jumps[-1]
                 cond_jump.false_target = VEXExprConverter.convert(irsb.next, manager)
 
             else:
@@ -718,12 +785,13 @@ class VEXIRSBConverter(Converter):
             statements.append(
                 Return(
                     manager.next_atom(),
-                    VEXExprConverter.convert(irsb.next, manager),
                     [],
                     ins_addr=manager.ins_addr,
                     vex_block_addr=manager.block_addr,
                     vex_stmt_idx=DEFAULT_STATEMENT,
                 )
             )
+        else:
+            raise NotImplementedError("Unsupported jumpkind")
 
         return Block(addr, irsb.size, statements=statements)
